﻿Imports System.Math
Imports System.Runtime.InteropServices


''' <summary>
''' It calculates the critical values for Dixon's ratios.
''' Properties: nDixon, iDixon, jDixon, etc. are defined in the manuscript  
''' </summary>
''' <remarks></remarks>
<ComVisible(True)> _
<ClassInterface(ClassInterfaceType.AutoDual)>
Public Class CriticalValues

    'Programmer : Mahendra P. Verma (mahendra@iie.org.mx)
    'Institution: Instituto de Investigaciones Electricas, Cuernavaca, Mexico
    '
    Const nAcc As Integer = 15      ' Abscissas Accuracy
    Const Eps As Double = 0.000000000000001
    Const b As Integer = 10         ' SBG polynomial 

#Region "Property definition"
    Private Updated, UpDatedPoly(3) As Boolean

    Private rVal(), rPr() As Double  ' Modify it

    Private xHerm(), xHHerm(), xLeg() As Double
    Private wHerm(), wHHerm(), wLeg() As Double

    Private nDix As Integer

    Public Property nDixon() As Integer
        Get
            Return nDix
        End Get
        Set(ByVal value As Integer)
            If Abs(value - nDix) > 0 Then
                nDix = value
                Updated = False
            End If
        End Set
    End Property

    Private iDix As Integer
    Public Property iDixon() As Integer
        Get
            Return iDix
        End Get
        Set(ByVal value As Integer)
            If Abs(value - iDix) > 0 Then
                iDix = value
                Updated = False
            End If
        End Set
    End Property

    Private jDix As Integer
    Public Property jDixon() As Integer
        Get
            Return jDix
        End Get
        Set(ByVal value As Integer)
            If Abs(value - jDix) > 0 Then
                jDix = value
                Updated = False
            End If
        End Set
    End Property

    Private nHerm As Integer = 30
    Public WriteOnly Property nHermPoly() As Integer
        Set(ByVal value As Integer)
            If Abs(nHerm - value) > 0 Then
                nHerm = value
                Updated = False
                UpDatedPoly(1) = False
            End If
        End Set
    End Property

    Private nHHerm As Integer = 15
    Public WriteOnly Property nHHermPoly() As Integer
        Set(ByVal value As Integer)
            If Abs(nHHerm - value) > 0 Then
                nHHerm = value
                Updated = False
                UpDatedPoly(2) = False
            End If
        End Set
    End Property

    Private hHermBall As Boolean
    Public WriteOnly Property HalfHermiteBall() As Boolean
        Set(ByVal value As Boolean)
            If value <> hHermBall Then
                hHermBall = value
                Updated = False
                UpDatedPoly(2) = False
            End If
        End Set
    End Property

    Private nLeg As Integer = 10
    Public WriteOnly Property nLegPoly() As Integer
        Set(ByVal value As Integer)
            If Abs(nLeg - value) > 0 Then
                nLeg = value
                Updated = False
                UpDatedPoly(3) = False
            End If
        End Set
    End Property

    Private nRAcc As Integer
    Public Property RAccuracy() As Integer
        Get
            Return nRAcc
        End Get
        Set(ByVal value As Integer)
            If Abs(nRAcc - value) > 0 Then
                Updated = False
                nRAcc = value
            End If
        End Set
    End Property

    Private mAlpha As Double
    Public Property AlphaValue() As Double
        Get
            Return mAlpha
        End Get
        Set(ByVal value As Double)
            If Abs(mAlpha - value) > Eps Then
                mAlpha = value
                Updated = False
            End If
        End Set
    End Property

    Private mrCritical As Double

    Public ReadOnly Property RCritical As Double
        Get
            If Updated = False Then Call Update()
            Return mrCritical
        End Get
    End Property

    Private RCtlProb As Double
    Public ReadOnly Property RProbablity As Double
        Get
            If Updated = False Then Call Update()
            Return RCtlProb
        End Get
    End Property

    Private RTotalProb As Double
    Public ReadOnly Property TotalProbability As Double
        Get
            If Updated = False Then Call Update()
            Return RTotalProb
        End Get
    End Property

#End Region


    ''' <summary>
    ''' It calculates the values of abscissa and weight of all the polynomials (n-Hermite, 
    ''' n-HHerm, and n-Legendre). All functions and subroutines associated with the polynomials
    ''' are called internally.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub AssignPoly()
        Dim i As Integer
        If UpDatedPoly(1) = False Then
            UpDatedPoly(1) = True
            ReDim xHerm(nHerm)
            ReDim wHerm(nHerm)
            Call FindXHermPoly(xHerm, nHerm, nAcc)
            For i = 1 To nHerm
                wHerm(i) = HermPolyW(nHerm, xHerm(i))
            Next
        End If

        If UpDatedPoly(2) = False Then
            UpDatedPoly(2) = True

            ReDim xHHerm(nHHerm)
            ReDim wHHerm(nHHerm)

            If hHermBall = True Then
                Call FindHHermPoly(nHHerm, 0, xHHerm, wHHerm, nAcc)
            Else
                Call FindXSBGPoly(xHHerm, nHHerm, nAcc, b)
                For i = 1 To nHHerm
                    wHHerm(i) = SBGPolyW(nHHerm, xHHerm(i), b)
                Next
            End If
        End If

        If UpDatedPoly(3) = False Then
            UpDatedPoly(3) = True

            ReDim xLeg(nLeg), wLeg(nLeg)

            Call FindXLegPoly(xLeg, nLeg, nAcc)
            For i = 1 To nLeg
                wLeg(i) = LegPolyW(nLeg, xLeg(i))
            Next
        End If
    End Sub

    ''' <summary>
    ''' It is main subroutine which is called internally when the value of any property
    ''' is changed. Similarly, it calls all the subroutines and functions.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub Update()

        Dim iter As Integer
        Dim R, R1, R2 As Double
        Dim RPrb, RPrb1, RPrb2 As Double
        Dim TestProb As Double

        If Updated = False Then
            Updated = True

            Call AssignPoly()

            RTotalProb = FindrProbLeg(1)

            TestProb = 1 - mAlpha

            If mAlpha < Eps Then
                RPrb = FindrProbLeg(1)
                mrCritical = 1
                RCtlProb = RPrb
            Else
                R1 = 0
                RPrb1 = 0

                R2 = 1
                RPrb2 = 1

                R = (R1 + R2) / 2
                RPrb = 1 - mAlpha
                RPrb = FindrProbLeg(R)
                iter = 0

                Do While Abs(RPrb - TestProb) > 10 ^ -nRAcc And iter < 200
                    iter = iter + 1
                    R = (R1 + R2) / 2
                    RPrb = 1 - mAlpha
                    RPrb = FindrProbLeg(R)
                    If RPrb > TestProb Then
                        R2 = R
                    Else
                        R1 = R
                    End If
                Loop
                mrCritical = R
                RCtlProb = RPrb
            End If
        End If
    End Sub

    ''' <summary>
    ''' It calculates the constant part of the equation 2 of the manuscript.
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="i"></param>
    ''' <param name="j"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function ConstN(ByVal n As Integer, ByVal i As Integer, ByVal j As Integer) As Double
        Dim sum As Double
        Dim k As Integer

        If n > 170 Then
            sum = 1
            If ((n - j - i - 1) >= i) And ((n - i - j - 1) >= j) Then
                For k = (n - j - i) To n
                    sum = sum * k
                Next
                sum = sum / Factorial(i - 1) / Factorial(j - 1)

            ElseIf (i >= (n - j - i - 1)) And (i >= j) Then
                For k = i + 1 To n
                    sum = sum * k
                Next
                sum = sum / Factorial(n - i - j - 1) / Factorial(j - 1)
            Else
                For k = j + 1 To n
                    sum = sum * k
                Next
                sum = sum / Factorial(i - 1) / Factorial(n - i - j - 1)
            End If
        Else
            sum = Factorial(n) / (Factorial(i - 1) * Factorial(n - j - i - 1) * Factorial(j - 1))
        End If

        Return sum
    End Function

    ''' <summary>
    '''  It calculates factorial n.
    ''' </summary>
    ''' <param name="n"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function Factorial(ByVal n As Integer) As Double
        Dim i As Integer

        Factorial = 1
        For i = 1 To n
            Factorial = Factorial * i
        Next
    End Function


    ''' <summary>
    ''' It calculates the probaility at r
    ''' </summary>
    ''' <param name="r"></param>
    ''' <param name="x"></param>
    ''' <param name="v"></param>
    ''' <param name="n"></param>
    ''' <param name="i"></param>
    ''' <param name="j"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function rProb(ByVal r As Double, ByVal x As Double, ByVal v As Double, ByVal n As Integer, ByVal i As Integer, ByVal j As Integer) As Double

        Dim NConst As Double
        Dim c1 As Double, c2 As Double, c3 As Double
        Dim g As Double

        'NConst = factorial(n) / (factorial(i - 1) * factorial(n - j - i - 1) * factorial(j - 1))

        NConst = ConstN(n, i, j)

        c1 = intPhi(x - v)
        c2 = intPhi(x)
        c3 = intPhi(x - r * v)

        g = 1.0#
        If Not ((i - 1) = 0) Then g = g * c1 ^ (i - 1)
        If Not ((n - j - i - 1) = 0) Then g = g * (c3 - c1) ^ (n - j - i - 1)
        If Not ((j - 1) = 0) Then g = g * (c2 - c3) ^ (j - 1)

        rProb = NConst * g * Phi(x - v) * Phi(x - r * v) * Phi(x)
    End Function


    ''' <summary>
    ''' It calculates the commulative probability density at r (r= 0 to 1)
    ''' </summary>
    ''' <param name="r"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function FindrProbLeg(ByVal r As Double) As Double

        Dim nH, nHHerm, nL As Integer

        Dim i, j, k As Integer
        Dim x, v, uCoef, tCoef, c1, c2, c3, g As Double

        Dim NConst, Sum, Sqr2Pi As Double

        'NConst = factorial(nDix) / (factorial(iDix - 1) * factorial(nDix - jDix - iDix - 1) * factorial(jDix - 1))

        NConst = ConstN(nDix, iDix, jDix)

        uCoef = Sqrt(2.0# / 3.0#)
        Sqr2Pi = (Sqrt(2 * PI)) ^ (3)

        nH = xHerm.GetUpperBound(0)
        nHHerm = xHHerm.GetUpperBound(0)
        nL = xLeg.GetUpperBound(0)

        ReDim rVal(nL)

        For i = 1 To nL
            rVal(i) = (xLeg(i) + 1.0) * (r / 2.0)
            tCoef = Sqrt(2.0# / (1.0# + rVal(i) * rVal(i)))

            Sum = 0.0
            For j = 1 To nH
                x = xHerm(j) * Sqrt(2.0 / 3.0)

                For k = 1 To nHHerm
                    v = xHHerm(k) * tCoef

                    c1 = intPhi(x - v)
                    c2 = intPhi(x)
                    c3 = intPhi(x - rVal(i) * v)

                    g = 1

                    If Not ((iDix - 1) = 0) Then g = g * c1 ^ (iDix - 1)
                    If Not ((nDix - jDix - iDix - 1) = 0) Then g = g * (c3 - c1) ^ (nDix - jDix - iDix - 1)
                    If Not ((jDix - 1) = 0) Then g = g * (c2 - c3) ^ (jDix - 1)

                    g = g * (Exp(xHerm(j) * xHHerm(k) * (1.0# + rVal(i)) * uCoef * tCoef))
                    g = g * wHerm(j) * wHHerm(k) * v * (r / 2.0)
                    Sum = Sum + g
                Next
            Next
            rVal(i) = Sum * NConst * uCoef * tCoef / Sqr2Pi
        Next

        Sum = 0
        For i = 1 To nL
            Sum = Sum + rVal(i) * wLeg(i)
        Next

        FindrProbLeg = Sum
    End Function


    ''' <summary>
    ''' The phi function (McBane, 2006)
    ''' </summary>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function Phi(ByVal x As Double) As Double
        Phi = 1 / Sqrt(2 * PI) * Exp(-x * x)
    End Function


    ''' <summary>
    ''' The integral of phi function (McBane, 2006)
    ''' </summary>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function intPhi(ByVal x As Double) As Double
        '
        '   integral -inf to x of phi(x)
        '
        Dim s As Double, t As Double, b As Double, q As Double, i As Integer

        If x >= -10 And x <= 10 Then
            s = x
            t = 0.0#
            b = x
            q = x * x
            i = 1
            Do While Not (s = t)
                t = s
                i = i + 2
                b = b * q / i
                s = s + b
            Loop
            intPhi = 0.5 + s * Exp(-0.5 * q - 0.918938533204673)
        Else
            If x <= -10 Then
                intPhi = 0
            Else
                intPhi = 1
            End If
        End If
    End Function

    ''' <summary>
    ''' Error function
    ''' </summary>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function erf(ByVal x As Double) As Double
        Dim sqr2 As Double

        sqr2 = Sqrt(2)
        erf = 2 * (-0.5 + intPhi(sqr2 * x))
    End Function


    ''' <summary>
    ''' To define the parameter n is odd or even
    ''' </summary>
    ''' <param name="n"> </param>
    ''' <returns> true, if n is even.</returns>
    ''' <remarks></remarks>
    Private Function OddEven(ByVal n As Integer) As Boolean

        Dim x As Integer

        x = CInt(n / 2)
        If Abs(n - x * 2) > 0 Then
            OddEven = True
        Else
            OddEven = False
        End If
    End Function

    ''' <summary>
    ''' It determines the n-Hermite polynomial at x. 
    ''' </summary>
    ''' <param name="n">degree of polynomial</param>
    ''' <param name="x">at the position x</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function HermPoly(ByVal n As Integer, ByVal x As Double) As Double
        Dim H() As Double

        ReDim H(n)
        Dim i As Integer
        H(0) = 1
        H(1) = 2 * x

        For i = 2 To n
            H(i) = 2 * x * H(i - 1) - 2 * (i - 1) * H(i - 2)
        Next

        HermPoly = H(n)
    End Function


    ''' <summary>
    ''' It calculates the weight of n-Hermite Polinomial at abscissa x
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function HermPolyW(ByVal n As Integer, ByVal x As Double) As Double
        Dim HermTemp As Double
        HermTemp = HermPoly(n - 1, x)
        HermPolyW = (2 ^ (n - 1) * Factorial(n) * Sqrt(PI)) / (n * n * HermTemp * HermTemp)
    End Function


    ''' <summary>
    ''' It calculates the first derivative of n-Hermite polynomial
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function HermPolyD(ByVal n As Integer, ByVal x As Double) As Double
        HermPolyD = 2 * n * HermPoly(n - 1, x)
    End Function


    ''' <summary>
    ''' It finds the values of abscissa on n-Hermite Polynomial with nAcc digits accuracy.
    ''' </summary>
    ''' <param name="XHermPoly"> the calculated values of n-abscissas of n-Hermite polynomial</param>
    ''' <param name="n"></param>
    ''' <param name="nAcc"></param>
    ''' <remarks></remarks>
    Private Sub FindXHermPoly(ByRef XHermPoly() As Double, ByVal n As Integer, ByVal nAcc As Integer)

        Dim Iter As Integer
        Dim nRoots As Integer, NRoot As Integer

        Dim f As Double, f1 As Double, f2 As Double
        Dim x As Double, X1 As Double, X2 As Double, Xinc As Double

        Dim XTemp As Double, FTemp As Double
        Dim i As Integer
        Dim xhrm() As Double

        ReDim XHermPoly(n)

        If OddEven(n) = False Then
            nRoots = CInt(n / 2)

        Else
            nRoots = CInt((n - 1) / 2)
        End If

        ReDim xhrm(nRoots)
        Xinc = 0.1
        x = x + Xinc

        X1 = x
        f = HermPoly(n, X1)
        f1 = f

        Do While NRoot < nRoots
            x = x + Xinc
            f = HermPoly(n, x)
            X2 = x
            f2 = f
            If f1 * f2 < 0 Then
                'Find root by the bisection method
                Iter = 0
                Do While Abs(X1 - X2) > 10 ^ -nAcc And Iter < 100
                    Iter = Iter + 1
                    XTemp = (X1 + X2) / 2
                    FTemp = HermPoly(n, XTemp)

                    If f1 * FTemp >= 0 Then
                        f1 = FTemp
                        X1 = XTemp
                    Else
                        f2 = FTemp
                        X2 = XTemp
                    End If
                Loop
                NRoot = NRoot + 1
                xhrm(NRoot) = XTemp
            End If
            X1 = X2
            f1 = f2
        Loop

        If OddEven(n) = True Then
            XHermPoly(NRoot + 1) = 0.0#
            For i = 1 To nRoots
                XHermPoly(nRoots + 1 + i) = xhrm(i)
            Next
        Else
            For i = 1 To nRoots
                XHermPoly(nRoots + i) = xhrm(i)
            Next

        End If
        For i = 1 To nRoots
            XHermPoly(i) = -xhrm(nRoots + 1 - i)
        Next
    End Sub

#Region "half Hermite"

    ''' <summary>
    ''' It calcuates the abssica and weight for half-range Hermite polynomials using the algorithm 
    ''' presented by James A. Ball (2003). Half-range generalized Hermite Polynomials and the 
    ''' related Gaussian quadratures. SIAM J. Numer. Anal. 40(6), 2311-2317. 
    ''' 
    ''' </summary>
    ''' <param name="n"> Input: degree of n-Hermite polynomial</param>
    ''' <param name="gamma">Input: real value greater than -1. 
    ''' In this work it is zero. Look Ball (2003) for detials.</param>
    ''' <param name="XPoly">Output: n-abssica in ascending order</param>
    ''' <param name="WPoly">Output: n-weights correspoding to the abssica.</param>
    ''' <remarks></remarks>
    Private Sub FindHHermPoly(ByVal n As Integer, ByVal gamma As Double,
                           ByRef XPoly() As Double, ByRef WPoly() As Double, ByVal nAcc As Integer)

        Dim gVec(n), dgVec(n), aCoeff(n), bCoeff(n), cCoeff(n), F(n) As Double
        Dim i As Integer, iter As Integer

        ReDim XPoly(n)
        ReDim WPoly(n)

        gVec = gVecInitial(n + 1, gamma)
        F = FVec(n, gamma, gVec)

        Dim sumF As Double
        sumF = 0
        For i = 1 To n
            sumF = sumF + F(i) * F(i)
        Next
        iter = 0
        Do While sumF > 1 * 10 ^ -nAcc And iter < 100
            iter = iter + 1
            For i = 1 To n
                ' calculate aCoeff
                If i = 1 Then
                    aCoeff(i) = 0
                Else
                    aCoeff(i) = DFiDgim(n, i, gamma, gVec)
                End If

                'calcuate bCoeff
                bCoeff(i) = DFiDgi(n, i, gamma, gVec)

                'calculate cCoeff
                If i = n Then
                    cCoeff(i) = 0
                Else
                    cCoeff(i) = DFiDgip(n, i, gamma, gVec)
                End If
            Next
            Call SolveTDMA(n, aCoeff, bCoeff, cCoeff, F, dgVec)

            For i = 1 To n
                gVec(i) = gVec(i) - dgVec(i)
            Next
            F = FVec(n, gamma, gVec)

            sumF = 0
            For i = 0 To n
                sumF = sumF + F(i) * F(i)
            Next
        Loop

        ' Calculate xPoly, wPloy
        Dim alpha(n), Beta(n) As Double

        alpha(0) = Exp(GammaLn(1 + gamma / 2)) / Exp(GammaLn((1 + gamma) / 2))
        Beta(0) = Sqrt(PI) / 2
        For i = 1 To n - 1
            alpha(i) = Sqrt((2 * i + gamma + 1) / 3 - gVec(i + 1) - gVec(i))
            Beta(i) = (i + gamma / 2) / 6 + gVec(i)
        Next


        Dim AMat(,), x(), xVec(,) As Double

        Dim nHH As Integer

        nHH = alpha.GetUpperBound(0)


        ReDim AMat(nHH, nHH)
        ReDim x(nHH)
        ReDim xVec(nHH, nHH)

        Dim Dvec() As Double
        Dim eVec() As Double

        ReDim Dvec(nHH)
        ReDim eVec(nHH)


        For i = 1 To nHH
            Dvec(i) = alpha(i - 1)

            If i = nHH Then
                eVec(i) = 0.0
            Else
                eVec(i) = Sqrt(Beta(i))
            End If
        Next

        Dim zero(5), weight(5), e(5) As Double

        Dim alfGauss() As Double, betaGauss() As Double

        ReDim alfGauss(nHH)
        ReDim betaGauss(nHH)

        For i = 1 To nHH
            alfGauss(i) = alpha(i - 1)
            betaGauss(i) = Beta(i - 1)
        Next
        tqli(Dvec, eVec, nHH, xVec)

        Dim wt(nHH) As Double
        Dim testsum(nHH) As Double

        For j = 1 To nHH
            testsum(j) = 0
            For i = 1 To nHH
                testsum(j) += xVec(i, j) * xVec(i, j)
            Next
            For i = 1 To nHH
                xVec(i, j) = xVec(i, j) / testsum(j)
            Next

        Next

        For i = 1 To nHH
            wt(i) = Beta(0) * xVec(1, i) * xVec(1, i)
        Next
        For i = 1 To nHH
            XPoly(i) = Dvec(i)
            WPoly(i) = wt(i)
        Next
    End Sub


    ''' <summary>
    ''' It calculates the eigenvalues and eigenvectors of tridiagonal symmetric matrix.
    ''' Numerical Recipes in C: The art of scientific computing 
    ''' William H. Press, Saul A. Teukolsky, William T. Vetterling, Brain P. Flannery
    ''' Second Edition
    ''' Cambridge University Press (1988)
    ''' </summary>
    ''' <param name="d">Input: d-vector diagonal elements. 1 to n.
    ''' The eigenvalues are return in d-vector in ascending order.</param>
    ''' <param name="e">Input: e-vector off-diagonal elements. 1 to n-1.
    ''' The return values are less than the accuracy of eigenvalues.
    ''' It can be used for the checking purposes. </param>
    ''' <param name="n">input: number of diagonal elements.</param>
    ''' <param name="z">Output: It is a nxn matrix whose columns contain the corresponding 
    ''' eigenvectors.</param>
    ''' <remarks></remarks>
    Private Sub tqli(ByRef d() As Double, ByRef e() As Double, ByVal n As Integer, ByRef z(,) As Double)

        Dim i, iter, k, l, m As Integer
        Dim b, c, dd, f, g, p, r, s As Double
        Dim Eps As Double = 0.00000000000001

        ReDim z(n, n)

        For i = 1 To n
            z(i, i) = 1.0
        Next
        For l = 1 To n
            iter = 0
            Do
                For m = l To n - 1
                    dd = Abs(d(m)) + Abs(d(m + 1))
                    If Abs(Abs(e(m)) + dd - dd) < Eps Then
                        Exit For
                    End If
                Next

                If m <> l Then
                    iter = iter + 1
                    If iter > 100 Then
                        Exit Sub
                    End If
                    g = (d(l + 1) - d(l)) / (2.0 * e(l))
                    r = pythag(g, 1.0)
                    g = d(m) - d(l) + e(l) / (g + sign(r, g))
                    s = 1.0
                    c = 1.0
                    p = 0.0
                    For i = m - 1 To l Step -1
                        f = s * e(i)
                        b = c * e(i)
                        r = pythag(f, g)
                        e(i + 1) = r
                        If r = 0.0 Then
                            d(i + 1) -= p
                            e(m) = 0.0
                            Exit For
                        End If
                        s = f / r
                        c = g / r
                        g = d(i + 1) - p
                        r = (d(i) - g) * s + 2.0 * c * b
                        p = s * r
                        d(i + 1) = g + p
                        g = c * r - b

                        ' calculate eigenvectors
                        For k = 1 To n
                            f = z(k, i + 1)
                            z(k, i + 1) = s * z(k, i) + c * f
                            z(k, i) = c * z(k, i) - s * f
                        Next
                    Next

                    If r = 0.0 AndAlso i >= 1 Then
                        Continue Do
                    End If
                    d(l) -= p
                    e(l) = g
                    e(m) = 0.0
                End If
            Loop While m <> l
        Next
        swap(d, z)
    End Sub


    ''' <summary>
    ''' It is used to swap the eigenvalues and eigenvectors in ascending order.
    ''' </summary>
    ''' <param name="e">Input and Output: Eigenvalues</param>
    ''' <param name="emat">Input and Output: Eigenvectors</param>
    ''' <remarks></remarks>
    Private Sub swap(ByRef e() As Double, ByRef emat(,) As Double)
        Dim n, k As Integer
        n = e.GetUpperBound(0)
        Dim m, l As Integer

        For k = 1 To n - 1
            m = k
            For l = k + 1 To n
                If e(l) < e(m) Then
                    m = l
                End If
            Next
            If Not k = m Then
                Dim temp As Double
                temp = e(k)
                e(k) = e(m)
                e(m) = temp
                For i = 1 To n
                    temp = emat(k, i)
                    emat(i, k) = emat(i, m)
                    emat(i, m) = temp
                Next
            End If
        Next
    End Sub

    Private Function sign(ByVal a As Double, ByVal b As Double) As Double
        If b > 0 Then
            Return Abs(a)
        Else
            Return -Abs(a)
        End If
    End Function

    Private Function pythag(ByVal a As Double, ByVal b As Double) As Double
        'Computes (a*a+b*b)^1/2 without destructive underflow or overflow
        Dim absa As Double = Abs(a)
        Dim absb As Double = Abs(b)
        If absa > absb Then
            Return absa * Sqrt(1.0 + (absb / absa) * (absb / absa))

        Else
            Return absb * Sqrt(1.0 + (absa / absb) * (absa / absb))
        End If

    End Function


    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="n"> number of equations </param>
    ''' <param name="a">sub-diagonal (means it is the diagonal below the main diagonal) -- indexed from 1..n, 
    ''' a(1)=0</param>
    ''' <param name="b">the main diagonal</param>
    ''' <param name="c">sup-diagonal (means it is the diagonal above the main diagonal) -- indexed from 1..n, 
    '''  c(n)=0 </param>
    ''' <param name="d">right part-- indexed from i.. n</param>
    ''' <param name="x">the answer-- indexed from i.. n</param>
    ''' 
    ''' http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm 
    ''' 
    ''' <remarks></remarks>
    Private Sub SolveTDMA(ByVal n As Integer, ByVal a() As Double, ByVal b() As Double,
                          ByVal c() As Double, ByVal d() As Double, ByRef x() As Double)
        Dim i As Integer

        For i = 1 To n
            If i = 1 Then
                c(i) = c(i) / b(i)
                d(i) = d(i) / b(i)
            Else
                Dim deno As Double
                deno = (b(i) - c(i - 1) * a(i))
                c(i) = c(i) / deno
                d(i) = (d(i) - d(i - 1) * a(i)) / deno
            End If
        Next

        'back substitution
        For i = n To 1 Step -1
            If i = n Then
                x(i) = d(i)
            Else
                x(i) = d(i) - c(i) * x(i + 1)
            End If
        Next
    End Sub


    ''' <summary>
    '''  Details are in Ball(2003)
    ''' </summary>
    ''' <param name="n"> Input: degree of half-range Hermite polynomial </param>
    ''' <param name="gamma">Input: gamma=0, in our program.</param>
    ''' <param name="g">initial guess values</param>
    ''' <returns>F-vector</returns>
    ''' <remarks></remarks>
    Private Function FVec(ByVal n As Integer, ByVal gamma As Double, ByVal g() As Double) As Double()

        Dim i As Integer
        Dim FTemp(n) As Double
        Dim y As Double

        For i = 1 To n
            Dim gim, gi, gip As Double
            y = 2 * i + gamma
            gim = g(i - 1)
            gi = g(i)
            gip = g(i + 1)
            FTemp(i) = ((y + 1) / 3 - gip - gi) * ((y - 1) / 3 - gi - gim) * (y / 12 + gi) ^ 2
            FTemp(i) = FTemp(i) - ((y / 6 - gi) ^ 2 - gamma * gamma / 16) ^ 2
        Next

        FVec = FTemp
    End Function


    ''' <summary>
    ''' It calculates the derivative of i th component of F-vector 
    ''' with i th value of g-vector (inital guess values)
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="i"></param>
    ''' <param name="gamma"></param>
    ''' <param name="g"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function DFiDgi(ByVal n As Integer, ByVal i As Integer, ByVal gamma As Double,
                            ByVal g() As Double) As Double
        Dim y As Double
        Dim gim, gi, gip As Double

        y = 2 * i + gamma
        If i = 0 Then
            gim = 0
            gi = g(i)
            gip = g(i + 1)
        ElseIf i = n Then
            gim = g(i - 1)
            gi = g(i)
            gip = 0
        Else
            gim = g(i - 1)
            gi = g(i)
            gip = g(i + 1)
        End If

        DFiDgi = ((y + 1) / 3 - gip - gi) * ((y - 1) / 3 - gi - gim) * 2 * (y / 12 + gi)
        DFiDgi = DFiDgi + ((y + 1) / 3 - gip - gi) * (-1) * (y / 12 + gi) ^ 2
        DFiDgi = DFiDgi + (-1) * ((y - 1) / 3 - gi - gim) * (y / 12 + gi) ^ 2

        DFiDgi = DFiDgi - 2 * ((y / 6 - gi) ^ 2 - gamma * gamma / 16) * 2 * (y / 6 - gi) * (-1)
    End Function


    ''' <summary>
    ''' It calculates the derivative of i th component of F-vector 
    ''' with i-1 th value of g-vector (inital guess values)
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="i"></param>
    ''' <param name="gamma"></param>
    ''' <param name="g"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function DFiDgim(ByVal n As Integer, ByVal i As Integer, ByVal gamma As Double,
                             ByVal g() As Double) As Double
        Dim y As Double
        Dim gim, gi, gip As Double

        y = 2 * i + gamma
        gim = g(i - 1)
        gi = g(i)
        gip = g(i + 1)
        DFiDgim = ((y + 1) / 3 - gip - gi) * (-1) * (y / 12 + gi) ^ 2
    End Function


    ''' <summary>
    ''' It calculates the derivative of i th component of F-vector 
    ''' with i+1 th value of g-vector (inital guess values)
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="i"></param>
    ''' <param name="gamma"></param>
    ''' <param name="g"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function DFiDgip(ByVal n As Integer, ByVal i As Integer, ByVal gamma As Double,
                             ByVal g() As Double) As Double
        Dim y As Double
        Dim gim, gi, gip As Double

        y = 2 * i + gamma
        If i = 0 Then
            gim = 0
            gi = g(i)
            gip = g(i + 1)
        ElseIf i = n Then
            gim = g(i - 1)
            gi = g(i)
            gip = 0
        Else
            gim = g(i - 1)
            gi = g(i)
            gip = g(i + 1)
        End If

        DFiDgip = (-1) * ((y - 1) / 3 - gi - gim) * (y / 12 + gi) ^ 2
    End Function



    ''' <summary>
    ''' It calculates the initical values of g, defined by Ball(2003).
    ''' </summary>
    ''' <param name="n"> Input: If there is need to calculate m half hermite polynomials. 
    ''' Then n should be m+1. </param>
    ''' <param name="gamma">Input: in our program it is equal to zero. </param>
    ''' <returns>Initial g values (0 to n). </returns>
    ''' <remarks></remarks>
    Private Function gVecInitial(ByVal n As Integer, ByVal gamma As Double) As Double()
        Dim C0, C1, C2, C3 As Double

        Dim i As Integer

        Dim gTemp(n) As Double
        Dim gamma2, gamma4, gamma6, gamma8 As Double

        gamma2 = gamma * gamma
        gamma4 = gamma2 * gamma2
        gamma6 = gamma4 * gamma2
        gamma8 = gamma4 * gamma4

        C0 = 1.0 / 36.0 - gamma2 / 8.0
        C1 = 23.0 / 432.0 - 11.0 / 48.0 * gamma2 + 3.0 / 32.0 * gamma2
        C2 = 1189.0 / 2592.0 - 409.0 / 192.0 * gamma2
        C2 = C2 + 75.0 / 64.0 * gamma4 + 9.0 / 64.0 * gamma6
        C3 = 196057.0 / 20736.0 - 153559.0 / 3456.0 * gamma2
        C3 = C3 + 7111.0 / 256.0 * gamma4 + 639.0 / 128.0 * gamma6 + 135.0 / 512.0 * gamma8

        For i = 0 To n
            If i = 0 Then
                Dim g0, g1 As Double
                Dim y As Double
                g0 = Exp(GammaLn(1 + gamma / 2)) / Exp(GammaLn((1 + gamma) / 2))

                y = 2 * 1 + gamma
                g1 = C0 / y + C1 / y ^ 3 + C2 / y ^ 5 + C3 / y ^ 7
                g0 = g0 * g0 + g1 - 1 / 3
                gTemp(0) = 0
            Else
                Dim y As Double
                y = 2 * i + gamma
                gTemp(i) = C0 / y + C1 / y ^ 3 + C2 / y ^ 5 + C3 / y ^ 7
            End If
        Next
        gVecInitial = gTemp
    End Function



    ''' <summary>
    ''' It calculates the natural logrithmic value of Gamma Function using the Lanczos method. 
    ''' C. Lanczos. A precision approximation of the gamma function. J.
    ''' Soc. Indust. Appl. Math. Ser. B Numer. Anal., 1:86{96, 1964.
    ''' </summary>
    ''' <param name="xx"> The real value to calculate Gamma function value.</param>
    ''' <returns></returns>
    ''' <remarks>The alorithm is taken from 
    ''' Numerical Recipes in C: The art of scientific computing. 
    ''' Second edition
    ''' William H. Press, Saul A. Teukolsky, William T. Vetterling, Brain P. Plannery
    ''' Cambridge University Press,
    ''' 1992</remarks>
    Public Shared Function GammaLn(ByVal xx As Double) As Double

        Dim x, y, tmp, ser As Double

        Static cof() As Double = {76.180091729471457, -86.505320329416776,
                                  24.014098240830911, -1.231739572450155,
                                   0.001208650973866179, -0.00000539523938495}
        Dim j As Integer
        x = xx
        y = x
        tmp = x + 5.5
        tmp -= (x + 0.5) * Log(tmp)
        ser = 1.0000000001900149

        For j = 0 To 5
            y = y + 1
            ser += cof(j) / y
        Next
        Return -tmp + Log(2.5066282746310007 * ser / x)
    End Function

#End Region


    ''' <summary>
    ''' The value of n-Legendre polynomial at x.
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function LegPoly(ByVal n As Integer, ByVal x As Double) As Double

        Dim Leg() As Double
        Dim i As Integer

        ReDim Leg(n)
        Leg(0) = 1
        Leg(1) = x

        For i = 2 To n
            Leg(i) = ((2 * i - 1) * x * Leg(i - 1) - (i - 1) * Leg(i - 2)) / i
        Next
        LegPoly = Leg(n)
    End Function
    ''' <summary>
    ''' The first derivate of n-Legendre polynomial
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function DLegPoly(ByVal n As Integer, ByVal x As Double) As Double

        DLegPoly = n * (LegPoly(n - 1, x) - x * LegPoly(n, x)) / (1 - x * x)
    End Function
    ''' <summary>
    ''' It calculates the weight of n-Legendre Polynomial at abscissa x.
    ''' </summary>
    ''' <param name="n"></param>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function LegPolyW(ByVal n As Integer, ByVal x As Double) As Double
        Dim DLegPolTemp As Double

        DLegPolTemp = DLegPoly(n, x)

        LegPolyW = 2 / ((1 - x * x) * DLegPolTemp * DLegPolTemp)
    End Function
    ''' <summary>
    ''' It calculates the abscissia of n-Legendre Polynomial with nAcc decimal digits accuracy
    ''' </summary>
    ''' <param name="XLegPoly">The calculated of values abscissas of n-Legendre polynomial</param>
    ''' <param name="n"></param>
    ''' <param name="nAcc"></param>
    ''' <remarks></remarks>
    Private Sub FindXLegPoly(ByRef XLegPoly() As Double, ByVal n As Integer, ByVal nAcc As Integer)

        Dim F As Double, f1 As Double, f2 As Double
        Dim Npnt As Integer
        Dim x As Double, X1 As Double, X2 As Double, Xinc As Double
        Dim XTemp As Double, FTemp As Double
        Dim NRoot As Integer, nRoots As Integer
        Dim Iter As Integer
        Dim i As Integer
        Dim XLeg() As Double

        ReDim XLegPoly(n)

        If OddEven(n) = False Then
            nRoots = CInt(n / 2)
        Else
            nRoots = CInt((n - 1) / 2)
        End If
        ReDim XLeg(nRoots)

        Npnt = n * 10
        Xinc = 2 / Npnt

        x = x + Xinc

        X1 = x
        F = LegPoly(n, X1)
        f1 = F

        Do While NRoot < nRoots
            x = x + Xinc
            F = LegPoly(n, x)
            X2 = x
            f2 = F
            If f1 * f2 < 0 Then
                'Find root by the bisection method
                Iter = 0
                Do While Abs(X1 - X2) > 10 ^ -nAcc And Iter < 100
                    Iter = Iter + 1
                    XTemp = (X1 + X2) / 2
                    FTemp = LegPoly(n, XTemp)

                    If f1 * FTemp >= 0 Then
                        f1 = FTemp
                        X1 = XTemp
                    Else
                        f2 = FTemp
                        X2 = XTemp
                    End If
                Loop
                NRoot = NRoot + 1
                XLeg(NRoot) = XTemp

            End If
            X1 = X2
            f1 = f2
        Loop

        If OddEven(n) = True Then
            XLegPoly(NRoot + 1) = 0.0#
            For i = 1 To nRoots
                XLegPoly(nRoots + 1 + i) = XLeg(i)
            Next
        Else
            For i = 1 To nRoots
                XLegPoly(nRoots + i) = XLeg(i)
            Next

        End If
        For i = 1 To nRoots
            XLegPoly(i) = -XLeg(nRoots + 1 - i)
        Next
    End Sub

#Region "SBG"
    Private Function SBGPoly(ByVal n As Integer, ByVal x As Double, ByVal b As Integer) As Double

        Dim i As Integer
        Dim px() As Double, p0() As Double, pb() As Double
        Dim alpha() As Double, beta() As Double, gamma() As Double
        Dim erfB As Double, ExpB As Double
        Dim SqrPi As Double

        ReDim px(n), p0(n), pb(n)
        ReDim alpha(n), beta(n), gamma(n)

        p0(0) = 1
        pb(0) = 1
        px(0) = 1
        erfB = erf(b)
        ExpB = Exp(-b * b)
        SqrPi = Sqrt(PI)

        p0(1) = -(1 - ExpB) / (SqrPi * erfB)
        pb(1) = b - (1 - ExpB) / (SqrPi * erfB)
        px(1) = x - (1 - ExpB) / (SqrPi * erfB)

        gamma(0) = (SqrPi / 2.0#) * erfB

        For i = 2 To n
            gamma(i - 1) = 0.5 * (i - 1) * gamma(i - 2)
            gamma(i - 1) = gamma(i - 1) - 0.5 * (ExpB * pb(i - 1) * pb(i - 2) - p0(i - 1) * p0(i - 2))
            alpha(i - 1) = 0.5 / gamma(i - 1) * (ExpB * pb(i - 1) * pb(i - 1) - p0(i - 1) * p0(i - 1))
            beta(i - 1) = -gamma(i - 1) / gamma(i - 2)

            px(i) = (x + alpha(i - 1)) * px(i - 1) + beta(i - 1) * px(i - 2)

            p0(i) = (alpha(i - 1)) * p0(i - 1) + beta(i - 1) * p0(i - 2)
            pb(i) = (b + alpha(i - 1)) * pb(i - 1) + beta(i - 1) * pb(i - 2)
        Next i

        SBGPoly = px(n)
    End Function

    Private Function SBGGamma(ByVal n As Integer, ByVal b As Integer) As Double

        Dim i As Integer
        Dim p0() As Double, pb() As Double
        Dim alpha() As Double, beta() As Double, gamma() As Double
        Dim erfB As Double, ExpB As Double
        Dim SqrPi As Double

        ReDim p0(n), pb(n), alpha(n), beta(n), gamma(n)

        p0(0) = 1
        pb(0) = 1

        erfB = erf(b)
        ExpB = Exp(-b * b)
        SqrPi = Sqrt(PI)


        p0(1) = -(1 - ExpB) / (SqrPi * erfB)
        pb(1) = b - (1 - ExpB) / (SqrPi * erfB)

        gamma(0) = (SqrPi / 2.0#) * erfB

        For i = 2 To n
            gamma(i - 1) = 0.5 * (i - 1) * gamma(i - 2)
            gamma(i - 1) = gamma(i - 1) - 0.5 * (ExpB * pb(i - 1) * pb(i - 2) - p0(i - 1) * p0(i - 2))
            alpha(i - 1) = 0.5 / gamma(i - 1) * (ExpB * pb(i - 1) * pb(i - 1) - p0(i - 1) * p0(i - 1))
            beta(i - 1) = -gamma(i - 1) / gamma(i - 2)
            p0(i) = (alpha(i - 1)) * p0(i - 1) + beta(i - 1) * p0(i - 2)
            pb(i) = (b + alpha(i - 1)) * pb(i - 1) + beta(i - 1) * pb(i - 2)
        Next i

        gamma(n) = 0.5 * (n) * gamma(n - 1)
        gamma(n) = gamma(n) - 0.5 * (ExpB * pb(n) * pb(n - 1) - p0(n) * p0(n - 1))
        SBGGamma = gamma(n)
    End Function

    Private Function SBGPolyD(ByVal n As Integer, ByVal x As Double, ByVal b As Integer) As Double

        Dim i As Integer

        Dim px() As Double, p0() As Double, pb() As Double
        Dim alpha() As Double, beta() As Double, gamma() As Double
        Dim erfB As Double, ExpB As Double
        Dim SqrPi As Double

        Dim pxD() As Double

        ReDim px(n), p0(n), pb(n), alpha(n), beta(n), gamma(n), pxD(n)

        pxD(0) = 0
        pxD(1) = 1

        p0(0) = 1
        pb(0) = 1
        erfB = erf(b)
        ExpB = Exp(-b * b)
        SqrPi = Sqrt(PI)

        p0(1) = -(1 - ExpB) / (SqrPi * erfB)
        pb(1) = b - (1 - ExpB) / (SqrPi * erfB)

        gamma(0) = (SqrPi / 2.0#) * erfB

        For i = 2 To n
            gamma(i - 1) = 0.5 * (i - 1) * gamma(i - 2)
            gamma(i - 1) = gamma(i - 1) - 0.5 * (ExpB * pb(i - 1) * pb(i - 2) - p0(i - 1) * p0(i - 2))
            alpha(i - 1) = 0.5 / gamma(i - 1) * (ExpB * pb(i - 1) * pb(i - 1) - p0(i - 1) * p0(i - 1))
            beta(i - 1) = -gamma(i - 1) / gamma(i - 2)

            pxD(i) = (x + alpha(i - 1)) * pxD(i - 1) + SBGPoly(i - 1, x, b) + beta(i - 1) * pxD(i - 2)

            p0(i) = (alpha(i - 1)) * p0(i - 1) + beta(i - 1) * p0(i - 2)
            pb(i) = (b + alpha(i - 1)) * pb(i - 1) + beta(i - 1) * pb(i - 2)
        Next i

        SBGPolyD = pxD(n)
    End Function

    Private Function SBGPolyW(ByVal n As Integer, ByVal x As Double, ByVal b As Double) As Double
        Dim gm, SBGP, SBGD As Double

        gm = SBGGamma(n - 1, b)
        SBGP = SBGPoly(n - 1, x, b)
        SBGD = SBGPolyD(n, x, b)
        SBGPolyW = gm / (SBGD * SBGP)

    End Function

    Private Sub FindXSBGPoly(ByRef XSBGPoly() As Double, ByVal n As Integer, ByVal nAcc As Integer, ByVal b As Integer)

        Dim Iter As Integer
        Dim nRoots As Integer, NRoot As Integer

        Dim f As Double, f1 As Double, f2 As Double
        Dim x As Double, X1 As Double, X2 As Double, Xinc As Double
        Dim XTemp As Double, FTemp As Double

        ReDim XSBGPoly(n)

        nRoots = n
        Xinc = 0.1
        x = x
        X1 = x
        f = SBGPoly(n, x, b)
        f1 = f

        Do While NRoot < nRoots
            x = x + Xinc
            f = SBGPoly(n, x, b)
            X2 = x
            f2 = f
            If f1 * f2 < 0 Then
                'Find root by the bisection method
                Iter = 0
                Do While Abs(X1 - X2) > 5 * 10 ^ -nAcc And Iter < 100
                    Iter = Iter + 1
                    XTemp = (X1 + X2) / 2
                    FTemp = SBGPoly(n, XTemp, b)

                    If f1 * FTemp >= 0 Then
                        f1 = FTemp
                        X1 = XTemp
                    Else
                        f2 = FTemp
                        X2 = XTemp
                    End If
                Loop
                NRoot = NRoot + 1
                XSBGPoly(NRoot) = XTemp

            End If
            X1 = X2
            f1 = f2
        Loop
    End Sub
#End Region

End Class


